from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

try:
    IXP_TRACKER_GEO_LOOKUP_FACTORY = str(settings.IXP_TRACKER_GEO_LOOKUP_FACTORY)
except AttributeError:
    IXP_TRACKER_GEO_LOOKUP_FACTORY = None
except(TypeError, ValueError):
    raise ImproperlyConfigured("IXP_TRACKER_GEO_LOOKUP_FACTORY must be a string value")

try:
    IXP_TRACKER_PEERING_DB_URL = str(settings.IXP_TRACKER_PEERING_DB_URL)
except AttributeError:
    IXP_TRACKER_PEERING_DB_URL = "https://www.peeringdb.com/api"
except(TypeError, ValueError):
    raise ImproperlyConfigured("IXP_TRACKER_PEERING_DB_URL must be a string value")

try:
    IXP_TRACKER_PEERING_DB_KEY = str(settings.IXP_TRACKER_PEERING_DB_KEY)
except AttributeError:
    IXP_TRACKER_PEERING_DB_KEY = None
except(TypeError, ValueError):
    raise ImproperlyConfigured("IXP_TRACKER_PEERING_DB_KEY must be a string value")

DATA_ARCHIVE_URL = "https://publicdata.caida.org/datasets/peeringdb/{year}/{month:02d}/peeringdb_2_dump_{year}_{month:02d}_{day:02d}.json"



# Generated by Django 5.1.7 on 2025-06-09 12:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0010_statspercountry_ixp_count'),
    ]

    operations = [
        migrations.AddField(
            model_name='statsperixp',
            name='rs_peering_rate',
            field=models.FloatField(default=None),
        ),
    ]



# Generated by Django 5.0.8 on 2024-09-19 07:58

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0008_alter_asn_id_alter_ixp_id_alter_ixpmember_id_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ixpmembershiprecord',
            name='member',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='ixp_tracker.ixpmember'),
        ),
    ]



from datetime import datetime, timedelta, timezone
import dateutil.parser

import pytest

from ixp_tracker import importers
from ixp_tracker.importers import ASNGeoLookup, dedupe_member_data
from ixp_tracker.models import IXPMember, IXPMembershipRecord
from tests.fixtures import create_asn_fixture, create_ixp_fixture

pytestmark = pytest.mark.django_db

dummy_member_data = {
    "asn": 12345,
    "ix_id": 2,
    "created": "2019-08-24T14:15:22Z",
    "updated": "2019-08-24T14:15:22Z",
    "is_rs_peer": True,
    "speed": 10000,
}

multiple_member_data = [
    {
        "asn": 56789,
        "ix_id": 5,
        "created": "2019-08-24T14:15:22Z",
        "updated": "2019-08-24T14:15:22Z",
        "is_rs_peer": False,
        "speed": 10000,
    },
    {
        "asn": 56789,
        "ix_id": 5,
        "created": "2018-08-24T14:15:22Z",
        "updated": "2018-08-24T14:15:22Z",
        "is_rs_peer": True,
        "speed": 3000,
    },
    {
        "asn": 56789,
        "ix_id": 5,
        "created": "2020-08-24T14:15:22Z",
        "updated": "2020-08-24T14:15:22Z",
        "is_rs_peer": False,
        "speed": 4000,
    }
]

date_now = datetime.now(timezone.utc)

class TestLookup(ASNGeoLookup):
    __test__ = False

    def __init__(self, default_status: str = "assigned"):
        self.default_status = default_status

    def get_iso2_country(self, asn: int, as_at: datetime) -> str:
        pass

    def get_status(self, asn: int, as_at: datetime) -> str:
        assert as_at <= datetime.now(timezone.utc)
        assert asn > 0
        return self.default_status


def test_with_no_data_does_nothing():
    processor = importers.process_member_data(date_now, TestLookup())
    processor([])

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_adds_new_member():
    create_asn_fixture(dummy_member_data["asn"])
    create_ixp_fixture(dummy_member_data["ix_id"])

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    members = IXPMember.objects.all()
    assert len(members) == 1
    current_membership = IXPMembershipRecord.objects.filter(member=members.first())
    assert len(current_membership) == 1


def test_does_nothing_if_no_asn_found():
    create_ixp_fixture(dummy_member_data["ix_id"])

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_does_nothing_if_no_ixp_found():
    create_asn_fixture(dummy_member_data["asn"])

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_updates_existing_member():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    member.save()

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    members = IXPMember.objects.all()
    assert len(members) == 1
    updated = members.first()
    assert updated.last_active.year > 2023


def test_updates_membership_for_existing_member():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=datetime(year=2023, month=7, day=13).date(),
        is_rs_peer=False,
        speed=500
    )
    membership.save()

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    membership = IXPMembershipRecord.objects.filter(member=member)
    assert len(membership) == 1
    current_membership = membership.first()
    assert current_membership.is_rs_peer
    assert current_membership.speed == 10000


def test_adds_new_membership_for_existing_member_marked_as_left():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=datetime(year=2018, month=1, day=3).date(),
        end_date=datetime(year=2018, month=7, day=13, tzinfo=timezone.utc),
        is_rs_peer=False,
        speed=500
    )
    membership.save()

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    members = IXPMember.objects.all()
    assert len(members) == 1
    current_membership = IXPMembershipRecord.objects.filter(member=member).order_by("-start_date")
    assert len(current_membership) == 2
    assert current_membership.first().end_date is None


def test_extends_membership_for_member_marked_as_left_if_created_before_date_left():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=datetime(year=2018, month=1, day=3).date(),
        end_date=datetime(year=2018, month=7, day=13, tzinfo=timezone.utc),
        is_rs_peer=False,
        speed=500
    )
    membership.save()

    member_data_with_created_date_before_date_left = dict(dummy_member_data)
    member_data_with_created_date_before_date_left["created"] = "2018-06-24T14:15:22Z"
    processor = importers.process_member_data(date_now, TestLookup())
    processor([member_data_with_created_date_before_date_left])

    members = IXPMember.objects.all()
    assert len(members) == 1
    current_membership = IXPMembershipRecord.objects.filter(member=member).order_by("-start_date")
    assert len(current_membership) == 1
    assert current_membership.first().end_date is None


def test_marks_member_as_left_that_is_no_longer_active():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    first_day_of_month = datetime.now(timezone.utc).replace(day=1)
    last_day_of_last_month = (first_day_of_month - timedelta(days=1))
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=last_day_of_last_month
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=dateutil.parser.isoparse(dummy_member_data["created"]).date(),
        is_rs_peer=False,
        speed=500
    )
    membership.save()
    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date is None

    processor = importers.process_member_data(date_now, TestLookup())
    processor([])

    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date.strftime("%Y-%m-%d") == last_day_of_last_month.strftime("%Y-%m-%d")


def test_does_not_mark_member_as_left_if_asn_is_assigned():
    asn = create_asn_fixture(dummy_member_data["asn"], "ZZ")
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime.now(timezone.utc)
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=dateutil.parser.isoparse(dummy_member_data["created"]).date(),
        is_rs_peer=False,
        speed=500
    )
    membership.save()
    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date is None

    processor = importers.process_member_data(date_now, TestLookup())
    processor([])

    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date is None


def test_marks_member_as_left_if_asn_is_not_assigned():
    asn = create_asn_fixture(dummy_member_data["asn"], "ZZ")
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    first_day_of_month = datetime.now(timezone.utc).replace(day=1)
    last_day_of_last_month = (first_day_of_month - timedelta(days=1))
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime.now(timezone.utc)
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=dateutil.parser.isoparse(dummy_member_data["created"]).date(),
        is_rs_peer=False,
        speed=500
    )
    membership.save()
    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date is None

    processor = importers.process_member_data(date_now, TestLookup("available"))
    processor([])

    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date.strftime("%Y-%m-%d") == last_day_of_last_month.strftime("%Y-%m-%d")


def test_does_not_mark_as_left_before_joining_date():
    asn = create_asn_fixture(dummy_member_data["asn"], "ZZ")
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    first_day_of_month = datetime.now(timezone.utc).replace(day=1)
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime.now(timezone.utc)
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=first_day_of_month.date(),
        is_rs_peer=False,
        speed=500
    )
    membership.save()

    processor = importers.process_member_data(date_now, TestLookup("available"))
    processor([])

    current_membership = IXPMembershipRecord.objects.filter(member=member)
    assert current_membership.first().end_date.strftime("%Y-%m-%d") == first_day_of_month.strftime("%Y-%m-%d")


def test_ensure_multiple_member_entries_does_not_trigger_multiple_new_memberships():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    member.save()
    # As we always create a new membership record if the most recent one has ended, for multiple ASN-IX combos this
    # could result in multiple new memberships being created
    membership = IXPMembershipRecord(
        member=member,
        start_date=datetime(year=2023, month=1, day=13, tzinfo=timezone.utc),
        is_rs_peer=False,
        speed=500,
        end_date=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    membership.save()

    member_data_with_created_date_after_date_left = dict(dummy_member_data)
    member_data_with_created_date_after_date_left["created"] = "2023-09-24T14:15:22Z"
    processor = importers.process_member_data(date_now, TestLookup())
    processor([member_data_with_created_date_after_date_left, member_data_with_created_date_after_date_left])

    memberships = IXPMembershipRecord.objects.filter(member=member)
    assert len(memberships) == 2


def test_do_not_add_new_membership_for_same_created_date():
    asn = create_asn_fixture(dummy_member_data["asn"])
    ixp = create_ixp_fixture(dummy_member_data["ix_id"])
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=dummy_member_data["updated"],
        last_active=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    member.save()
    # As we always create a new membership record if the most recent one has ended, for multiple ASN-IX combos this
    # could result in multiple new memberships being created
    membership = IXPMembershipRecord(
        member=member,
        start_date=dateutil.parser.isoparse(dummy_member_data["created"]).date(),
        is_rs_peer=False,
        speed=500,
        end_date=datetime(year=2023, month=7, day=13, tzinfo=timezone.utc)
    )
    membership.save()

    processor = importers.process_member_data(date_now, TestLookup())
    processor([dummy_member_data])

    memberships = IXPMembershipRecord.objects.filter(member=member)
    assert len(memberships) == 1


def test_dedupes_member_data_before_processing():
    deduped_data = dedupe_member_data(multiple_member_data)

    assert len(deduped_data) == 1


def test_set_rs_peer_to_true_if_any_member_is_set_to_true():
    deduped_data = dedupe_member_data(multiple_member_data)

    deduped_member = deduped_data[0]
    assert deduped_member["is_rs_peer"]


def test_speed_for_deduped_members_is_sum_of_all_speeds():
    deduped_data = dedupe_member_data(multiple_member_data)

    deduped_member = deduped_data[0]
    assert deduped_member["speed"] == 17000



# Generated by Django 5.1.7 on 2025-06-12 11:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0011_statsperixp_rs_peering_rate'),
    ]

    operations = [
        migrations.AddField(
            model_name='statspercountry',
            name='routed_asn_count',
            field=models.IntegerField(default=None, null=True),
        ),
        migrations.AddField(
            model_name='statspercountry',
            name='routed_asns_ixp_member_rate',
            field=models.FloatField(default=None, null=True),
        ),
        migrations.AddField(
            model_name='statsperixp',
            name='local_routed_asns_members_rate',
            field=models.FloatField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name='statsperixp',
            name='rs_peering_rate',
            field=models.FloatField(default=None, null=True),
        ),
    ]



import json
from datetime import datetime, timezone

import pytest
import responses

from django_test_app.settings import IXP_TRACKER_PEERING_DB_URL
from ixp_tracker import importers
from ixp_tracker.models import ASN

pytestmark = pytest.mark.django_db

dummy_asn_data = {
    "id": 3,
    "asn": 6543,
    "name": "New ASN",
    "info_type": "non-profit",
    "created": "2019-08-24T14:15:22Z",
    "updated": "2019-08-24T14:15:22Z",
}


class TestLookup:

    def get_iso2_country(self, asn: int, as_at: datetime) -> str:
        assert as_at <= datetime.now(timezone.utc)
        assert asn > 0
        return "AU"

    def get_status(self, asn: int, as_at: datetime) -> str:
        pass


def test_with_empty_response_does_nothing():
    processor = importers.process_asn_data(TestLookup())
    processor([])

    asns = ASN.objects.all()
    assert len(asns) == 0


def test_with_no_existing_data_gets_all_data():
    with responses.RequestsMock() as rsps:
        rsps.get(
            url=IXP_TRACKER_PEERING_DB_URL + "/net?limit=200&skip=0",
            body=""
        )
        importers.import_asns(TestLookup(), False)

        asns = ASN.objects.all()
        assert len(asns) == 0


def test_imports_new_asn():
    processor = importers.process_asn_data(TestLookup())
    processor([dummy_asn_data])

    asns = ASN.objects.all()
    assert len(asns) == 1


def test_updates_existing_data():
    asn = ASN(
        name="Network Org",
        number=dummy_asn_data["asn"],
        peeringdb_id=dummy_asn_data["id"],
        network_type="other",
        registration_country_code="CH",
        created=datetime(2019, 1, 1).replace(tzinfo=timezone.utc),
        last_updated=datetime(2024, 5, 1).replace(tzinfo=timezone.utc)
    )
    asn.save()
    with responses.RequestsMock() as rsps:
        rsps.get(
            url=IXP_TRACKER_PEERING_DB_URL + "/net?updated__gte=2024-05-01&limit=200&skip=0",
            body=json.dumps({"data": [dummy_asn_data]}),
        )
        rsps.get(
            url=IXP_TRACKER_PEERING_DB_URL + "/net?updated__gte=2024-05-01&limit=200&skip=200",
            body=json.dumps({"data": []}),
        )
        importers.import_asns(TestLookup(), False)

        asns = ASN.objects.all()
        assert len(asns) == 1
        updated = asns.filter(peeringdb_id=dummy_asn_data["id"]).first()
        assert updated.name == "New ASN"
        assert updated.registration_country_code == "AU"


def test_handles_errors_with_source_data():
    data_with_problems = dummy_asn_data
    data_with_problems["updated"] = "abc"
    data_with_problems["asn"] = "foobar"

    processor = importers.process_asn_data(TestLookup())
    processor([data_with_problems])

    asns = ASN.objects.all()
    assert len(asns) == 0



import pytest
from datetime import datetime, timedelta, timezone

from ixp_tracker.importers import toggle_ixp_active_status
from ixp_tracker.models import IXP
from tests.fixtures import create_ixp_fixture, create_member_fixture

pytestmark = pytest.mark.django_db
processing_date = datetime.now(timezone.utc)


def test_active_ixp_with_3_members_remains_active():
    ixp = create_ixp_fixture(1)
    create_member_fixture(ixp, 12345)
    create_member_fixture(ixp, 23456)
    create_member_fixture(ixp, 34567)

    toggle_ixp_active_status(processing_date)

    updated_ixp = IXP.objects.get(peeringdb_id=1)

    assert updated_ixp.active_status
    assert updated_ixp.last_updated == ixp.last_updated


def test_active_ixp_gone_from_three_to_two_active_members_is_marked_inactive():
    ixp = create_ixp_fixture(1)
    end_date = processing_date.replace(day=1) - timedelta(days=1)
    create_member_fixture(ixp, 12345)
    create_member_fixture(ixp, 23456)
    create_member_fixture(ixp, 34567, date_left=end_date)

    toggle_ixp_active_status(processing_date)

    updated_ixp = IXP.objects.get(peeringdb_id=1)

    assert updated_ixp.active_status is False
    assert updated_ixp.last_updated == processing_date


def test_inactive_ixp_with_two_active_members_remains_inactive():
    ixp = create_ixp_fixture(1)
    ixp.active_status = False
    ixp.save()
    end_date = processing_date.replace(day=1) - timedelta(days=1)
    create_member_fixture(ixp, 12345)
    create_member_fixture(ixp, 23456)
    create_member_fixture(ixp, 34567, date_left=end_date)

    toggle_ixp_active_status(processing_date)

    updated_ixp = IXP.objects.get(peeringdb_id=1)

    assert updated_ixp.active_status is False
    assert updated_ixp.last_updated == ixp.last_updated


def test_inactive_ixp_with_three_active_members_marked_active():
    ixp = create_ixp_fixture(1)
    ixp.active_status = False
    ixp.save()
    create_member_fixture(ixp, 12345)
    create_member_fixture(ixp, 23456)
    create_member_fixture(ixp, 34567)

    toggle_ixp_active_status(processing_date)

    updated_ixp = IXP.objects.get(peeringdb_id=1)

    assert updated_ixp.active_status
    assert updated_ixp.last_updated == processing_date



# Generated by Django 5.0.8 on 2024-09-16 13:56

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0004_ixpmember_ixpmember_unique_ixp_membership'),
    ]

    operations = [
        migrations.CreateModel(
            name='StatsPerCountry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('country', models.CharField(max_length=2)),
                ('stats_date', models.DateField()),
                ('asn_count', models.IntegerField()),
                ('member_count', models.IntegerField()),
                ('asns_ixp_member_rate', models.FloatField()),
                ('total_capacity', models.FloatField()),
            ],
            options={
                'verbose_name': 'Per-country stats',
            },
        ),
        migrations.CreateModel(
            name='StatsPerIXP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('stats_date', models.DateField()),
                ('capacity', models.FloatField()),
                ('members', models.IntegerField()),
                ('local_asns_members_rate', models.FloatField()),
            ],
            options={
                'verbose_name': 'IXP stats',
            },
        ),
        migrations.AddConstraint(
            model_name='statspercountry',
            constraint=models.UniqueConstraint(fields=('country', 'stats_date'), name='ixp_tracker_unique_per_country_stats'),
        ),
        migrations.AddField(
            model_name='statsperixp',
            name='ixp',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ixp_tracker.ixp'),
        ),
        migrations.AddConstraint(
            model_name='statsperixp',
            constraint=models.UniqueConstraint(fields=('ixp', 'stats_date'), name='ixp_tracker_unique_ixp_stats'),
        ),
    ]



# Generated by Django 5.0.8 on 2024-08-08 12:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ASN',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('number', models.IntegerField()),
                ('peeringdb_id', models.IntegerField(null=True)),
                ('network_type', models.CharField(choices=[('access', 'Access'), ('transit', 'Transit'), ('content', 'Content'), ('enterprise', 'Enterprise'), ('NREN', 'Research and Education'), ('non-profit', 'Non-profit'), ('not-disclosed', 'Not Disclosed'), ('other', 'Other')], default='not-disclosed', max_length=200)),
                ('registration_country', models.CharField(max_length=2)),
                ('created', models.DateTimeField()),
                ('last_updated', models.DateTimeField()),
            ],
            options={
                'verbose_name': 'AS Number',
                'verbose_name_plural': 'AS Numbers',
            },
        ),
        migrations.AddConstraint(
            model_name='asn',
            constraint=models.UniqueConstraint(fields=('number',), name='ixp_tracker_unique_as_number'),
        ),
    ]



from datetime import datetime, timedelta, timezone
from typing import List

import pytest

from ixp_tracker.models import StatsPerCountry
from ixp_tracker.stats import generate_stats
from tests.test_ixp_stats import create_member_fixture
from tests.test_members_import import create_ixp_fixture

pytestmark = pytest.mark.django_db


class MockLookup:

    def __init__(self, default_status: str = "assigned"):
        self.default_status = default_status

    def get_iso2_country(self, asn: int, as_at: datetime) -> str:
        pass

    def get_status(self, asn: int, as_at: datetime) -> str:
        pass

    def get_asns_for_country(self, country: str, as_at: datetime) -> List[int]:
        return [12345, 446, 789, 5050, 54321]

    def get_routed_asns_for_country(self, country: str, as_at: datetime) -> List[int]:
        return [12345, 446, 789, 54321]


def test_with_no_data_generates_no_stats():
    generate_stats(MockLookup())

    stats = StatsPerCountry.objects.all()
    assert len(stats) == 249
    first_stat = stats.first()
    assert first_stat.member_count == 0


def test_generates_stats():
    ixp_one = create_ixp_fixture(123, "CH")
    create_member_fixture(ixp_one, 12345, 500)
    create_member_fixture(ixp_one, 67890, 10000)
    create_member_fixture(ixp_one, 3456, 10000)
    ixp_two = create_ixp_fixture(124, "CH")
    create_member_fixture(ixp_two, 5050, 6000)
    create_member_fixture(ixp_two, 67890, 10000)
    create_member_fixture(ixp_two, 3456, 10000)

    generate_stats(MockLookup())

    stats = StatsPerCountry.objects.filter(country_code="CH").first()
    # The default fixture does not have a recent last_active date so technically they shouldn't be counted here
    assert stats.ixp_count == 2
    assert stats.asn_count == 5
    assert stats.routed_asn_count == 4
    assert stats.member_count == 4
    assert stats.total_capacity == 46.5
    assert stats.asns_ixp_member_rate == 0.4
    assert stats.routed_asns_ixp_member_rate == 0.25


def test_generates_ixp_counts():
    stats_date = (datetime.now(timezone.utc) - timedelta(weeks=16)).replace(day=1)
    one_month_before = (stats_date - timedelta(days=1)).replace(day=1)
    one_month_after = (stats_date + timedelta(days=35)).replace(day=1)
    # currently_active with three members
    active = create_ixp_fixture(123, "CH")
    create_member_fixture(active, 12345, 500, member_since=one_month_before, date_left=one_month_after)
    create_member_fixture(active, 345, 500, member_since=one_month_before, date_left=one_month_after)
    create_member_fixture(active, 9876, 500, member_since=one_month_before, date_left=one_month_after)
    # member active in the past
    member_in_past = create_ixp_fixture(124, "CH")
    create_member_fixture(member_in_past, 12345, 500, member_since=one_month_before, date_left=one_month_before)
    # member not yet active (as we are generating historical stats there could be members in the future)
    member_in_future = create_ixp_fixture(125, "CH")
    create_member_fixture(member_in_future, 12345, 500, member_since=one_month_after, date_left=None)
    # currently_active but only two members
    not_enough_members = create_ixp_fixture(126, "CH")
    create_member_fixture(not_enough_members, 8887, 500, member_since=one_month_before, date_left=one_month_after)
    create_member_fixture(not_enough_members, 7778, 500, member_since=one_month_before, date_left=one_month_after)

    generate_stats(MockLookup(), stats_date)

    stats = StatsPerCountry.objects.filter(country_code="CH").first()
    assert stats.ixp_count == 1
    assert stats.member_count == 3


def test_handles_invalid_country():
    create_ixp_fixture(123, "XK")

    generate_stats(MockLookup())

    country_stats = StatsPerCountry.objects.filter(country_code="XK").first()
    assert country_stats is None



import logging
from datetime import datetime, timezone
from typing import Dict, List, TypedDict, Union

from django_countries import countries
from django.db.models import Q

from ixp_tracker.importers import ASNGeoLookup, is_ixp_active
from ixp_tracker.models import IXP, IXPMember, IXPMembershipRecord, StatsPerCountry, StatsPerIXP

logger = logging.getLogger("ixp_tracker")


class CountryStats(TypedDict):
    ixp_count: int
    all_asns: Union[List[int], None]
    routed_asns: Union[List[int], None]
    member_asns: set
    total_capacity: int


def generate_stats(geo_lookup: ASNGeoLookup, stats_date: datetime = None):
    stats_date = stats_date or datetime.now(timezone.utc)
    stats_date = stats_date.replace(day=1)
    ixps = IXP.objects.filter(created__lte=stats_date).all()
    all_members = (IXPMember.objects
                   .filter(
                        Q(memberships__start_date__lte=stats_date) &
                        (Q(memberships__end_date=None) | Q(memberships__end_date__gte=stats_date))
                    )).all()
    all_stats_per_country: Dict[str, CountryStats] = {}
    for code, _ in list(countries):
        all_stats_per_country[code] = {
            "ixp_count": 0,
            "all_asns": None,
            "routed_asns": None,
            "member_asns": set(),
            "total_capacity": 0
        }
    for ixp in ixps:
        logger.debug("Calculating growth stats for IXP", extra={"ixp": ixp.id})
        members = [member for member in all_members if member.ixp == ixp]
        member_count = len(members)
        capacity = 0
        rs_peers = 0
        for member in members:
            # Make sure we get the relevant membership record for the stats_date
            membership = (IXPMembershipRecord.objects
                .filter(start_date__lte=stats_date, member=member)
                .filter(Q(end_date=None) | Q(end_date__gte=stats_date))).first()
            if membership is not None:
                capacity += membership.speed
                if membership.is_rs_peer:
                    rs_peers += 1
        ixp_country = ixp.country_code
        country_stats = all_stats_per_country.get(ixp_country)
        if country_stats is None:
            logger.warning("Country not found", extra={"country": ixp_country})
            country_stats = {
                "ixp_count": 0,
                "all_asns": None,
                "routed_asns": None,
                "member_asns": set(),
                "total_capacity": 0
            }
            all_stats_per_country[ixp_country] = country_stats
        if country_stats.get("all_asns") is None:
            all_stats_per_country[ixp_country]["all_asns"] = geo_lookup.get_asns_for_country(ixp_country, stats_date)
        if country_stats.get("routed_asns") is None:
            all_stats_per_country[ixp_country]["routed_asns"] = geo_lookup.get_routed_asns_for_country(ixp_country, stats_date)
        member_asns = [member.asn.number for member in members]
        local_asns_members_rate = calculate_local_asns_members_rate(member_asns, all_stats_per_country[ixp_country]["all_asns"])
        local_routed_asns_members_rate = calculate_local_asns_members_rate(member_asns, all_stats_per_country[ixp_country]["routed_asns"])
        rs_peering_rate = rs_peers / member_count if rs_peers else 0
        # We always save the stats per IXP so we can track stats across time (e.g. if an IXP becomes inactive then active again)
        StatsPerIXP.objects.update_or_create(
            ixp=ixp,
            stats_date=stats_date.date(),
            defaults={
                "ixp": ixp,
                "stats_date": stats_date.date(),
                "members": member_count,
                "capacity": (capacity/1000),
                "local_asns_members_rate": local_asns_members_rate,
                "local_routed_asns_members_rate": local_routed_asns_members_rate,
                "rs_peering_rate": rs_peering_rate,
            }
        )
        # Only aggregate this IXP's stats into the country stats if it's active
        if is_ixp_active(members):
            all_stats_per_country[ixp_country]["ixp_count"] += 1
            # We only count unique ASNs that are members of an IXP in a country
            all_stats_per_country[ixp_country]["member_asns"] |= set(member_asns)
            # But we count capacity for all members, i.e. an ASN member at 2 IXPs will have capacity at each included in the sum
            all_stats_per_country[ixp_country]["total_capacity"] += capacity
    for code, _ in list(countries):
        country_stats = all_stats_per_country[code]
        if country_stats.get("all_asns") is None:
            country_stats["all_asns"] = geo_lookup.get_asns_for_country(code, stats_date)
        if country_stats.get("routed_asns") is None:
            country_stats["routed_asns"] = geo_lookup.get_routed_asns_for_country(code, stats_date)
        local_asns_members_rate = calculate_local_asns_members_rate(country_stats["member_asns"], country_stats["all_asns"])
        local_routed_asns_members_rate = calculate_local_asns_members_rate(country_stats["member_asns"], country_stats["routed_asns"])
        StatsPerCountry.objects.update_or_create(
            country_code=code,
            stats_date=stats_date.date(),
            defaults={
                "ixp_count": country_stats["ixp_count"],
                "asn_count": len(country_stats["all_asns"]),
                "routed_asn_count": len(country_stats["routed_asns"]),
                "member_count": len(country_stats["member_asns"]),
                "asns_ixp_member_rate": local_asns_members_rate,
                "routed_asns_ixp_member_rate": local_routed_asns_members_rate,
                "total_capacity": (country_stats["total_capacity"]/1000),
            }
        )


def calculate_local_asns_members_rate(member_asns: List[int], country_asns: List[int]) -> float:
    if len(country_asns) == 0:
        return 0
    # Ignore the current country for a member ASN (as that might have changed) but just get all current members
    # that are in the list of ASNs registered to the country at the time
    members_in_country = [asn for asn in member_asns if asn in country_asns]
    return len(members_in_country) / len(country_asns)



# Generated by Django 5.0.8 on 2025-04-24 09:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0009_alter_ixpmembershiprecord_member'),
    ]

    operations = [
        migrations.AddField(
            model_name='statspercountry',
            name='ixp_count',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.0.8 on 2024-08-14 10:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ixp_tracker', '0003_alter_asn_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='IXPMember',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('member_since', models.DateField()),
                ('last_updated', models.DateTimeField()),
                ('is_rs_peer', models.BooleanField(default=False)),
                ('speed', models.IntegerField(null=True)),
                ('date_left', models.DateField(null=True)),
                ('last_active', models.DateTimeField(null=True)),
                ('asn', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ixp_tracker.asn')),
                ('ixp', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ixp_tracker.ixp')),
            ],
            options={
                'verbose_name': 'IXP Member',
                'verbose_name_plural': 'IXP Members',
            },
        ),
        migrations.AddConstraint(
            model_name='ixpmember',
            constraint=models.UniqueConstraint(fields=('ixp', 'asn'), name='ixp_tracker_unique_ixp_membership'),
        ),
    ]
